# -*- coding: utf-8 -*-

"""
@date: 2021/2/2 下午5:19
@file: repvgg.py
@author: zj
@description: RegVGG，参考[RepVGG/repvgg.py](https://github.com/DingXiaoH/RepVGG/blob/main/repvgg.py)
"""

from zcls.model import registry
from zcls.model.misc import load_pretrained_weights
from ..base_recognizer import BaseRecognizer

url_map = {
    'repvgg_a0_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_a0_train_imagenet_2063bcbc.pth",
    'repvgg_a0_infer': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_a0_infer_imagenet_7af195b9.pth",
    'repvgg_a1_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_a1_train_imagenet_c3def5af.pth",
    'repvgg_a1_infer': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_a1_infer_imagenet_1b9ae0cc.pth",
    'repvgg_a2_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_a2_train_imagenet_62d83c2a.pth",
    'repvgg_a2_infer': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_a2_infer_imagenet_6f627938.pth",
    'repvgg_b0_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b0_train_imagenet_69920b69.pth",
    'repvgg_b0_infer': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b0_infer_imagenet_d6d57b2d.pth",
    'repvgg_b1_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b1_train_imagenet_ea3e69be.pth",
    'repvgg_b1_infer': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b1_infer_imagenet_64bf1d63.pth",
    'repvgg_b1g2_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b1g2_train_imagenet_902d9fee.pth",
    'repvgg_b1g2_infer': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b1g2_infer_imagenet_66b82aab.pth",
    'repvgg_b1g4_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b1g4_train_imagenet_5d492d86.pth",
    'repvgg_b1g4_infer': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b1g4_infer_imagenet_405181ae.pth",
    'repvgg_b2_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b2_train_imagenet_7c25ed20.pth",
    'repvgg_b2_infer': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b2_infer_imagenet_9653c204.pth",
    'repvgg_b2g4_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b2g4_train_imagenet_d4e83c66.pth",
    'repvgg_b2g4_infer': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b2g4_infer_imagenet_da2eaa1e.pth",
    'repvgg_b3_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b3_train_imagenet_3e026a56.pth",
    'repvgg_b3_infer': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b3_infer_imagenet_02e67a97.pth",
    'repvgg_b3g4_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b3g4_train_imagenet_4f7773c1.pth",
    'repvgg_b3g4_infer': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b3g4_infer_imagenet_3719b8ab.pth",
    'repvgg_d2se_train': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/repvgg_b3g4_train_imagenet_4f7773c1.pth",
    'repvgg_d2se_infer': "",
}


@registry.RECOGNIZER.register('RepVGG')
class RepVGG(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)

    def _init_weights(self, cfg):
        pretrained_local = cfg.MODEL.RECOGNIZER.PRETRAINED_LOCAL
        pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
        num_classes = cfg.MODEL.HEAD.NUM_CLASSES

        model_name = cfg.MODEL.BACKBONE.ARCH
        assert isinstance(model_name, str)
        if model_name.startswith('repvgg'):
            if len(cfg.MODEL.CONV.ADD_BLOCKS) == 0:
                model_name += '_infer'
            else:
                model_name += '_train'
        load_pretrained_weights(self, model_name,
                                weights_path=None if pretrained_local == '' else pretrained_local,
                                load_fc=pretrained_num_classes == num_classes,
                                verbose=True,
                                url_map=url_map if cfg.MODEL.RECOGNIZER.PRETRAINED_REMOTE else None
                                )


# -*- coding: utf-8 -*-

"""
@date: 2020/12/24 下午7:38
@file: shufflenetv1.py
@author: zj
@description: 
"""

from zcls.model import registry
from zcls.model.misc import load_pretrained_weights
from ..base_recognizer import BaseRecognizer

"""
Note 1: Empirically g = 3 usually has a proper trade-off between accuracy and actual inference time
Note 2: Comparing ShuffleNet 2× with MobileNet whose complexity are comparable (524 vs. 569 MFLOPs)
"""

url_map = {
    'shufflenetv1_3g0_5x': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/shufflenetv1_3g0_5x_imagenet_de2ce242.pth",
    'shufflenetv1_3g1x': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/shufflenetv1_3g1x_imagenet_4b9454f2.pth",
    'shufflenetv1_3g1_5x': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/shufflenetv1_3g1_5x_imagenet_9ac247e1.pth",
    'shufflenetv1_3g2x': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/shufflenetv1_3g2x_imagenet_9835481a.pth",
    'shufflenetv1_8g0_5x': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/shufflenetv1_8g0_5x_imagenet_5940e6d3.pth",
    'shufflenetv1_8g1x': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/shufflenetv1_8g1x_imagenet_941bfdc6.pth",
    'shufflenetv1_8g1_5x': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/shufflenetv1_8g1_5x_imagenet_76b3f38d.pth",
    'shufflenetv1_8g2x': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/shufflenetv1_8g2x_imagenet_9fb71642.pth",
}


@registry.RECOGNIZER.register('ShuffleNetV1')
class ShuffleNetV1(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)

    def _init_weights(self, cfg):
        pretrained_local = cfg.MODEL.RECOGNIZER.PRETRAINED_LOCAL
        pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
        num_classes = cfg.MODEL.HEAD.NUM_CLASSES
        load_pretrained_weights(self, cfg.MODEL.BACKBONE.ARCH,
                                weights_path=None if pretrained_local == '' else pretrained_local,
                                load_fc=pretrained_num_classes == num_classes,
                                verbose=True,
                                url_map=url_map if cfg.MODEL.RECOGNIZER.PRETRAINED_REMOTE else None
                                )



# -*- coding: utf-8 -*-

"""
@date: 2021/10/4 下午1:26
@file: util.py
@author: zj
@description: 
"""

import cv2
from typing import Any

import numpy as np
from PIL import Image


def default_loader(path: str, rgb=False) -> Any:
    image = cv2.imread(path)
    if rgb:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image


def default_converter(img: Image.Image, rgb=False) -> Any:
    image = np.array(img)
    if not rgb:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    return image



# -*- coding: utf-8 -*-

"""
@date: 2020/10/19 上午10:03
@file: base_evaluator.py
@author: zj
@description: 
"""

from abc import ABCMeta, abstractmethod
import torch

from zcls.config.key_word import KEY_OUTPUT
from zcls.util.metrics import topk_accuracy


class BaseEvaluator(metaclass=ABCMeta):

    def __init__(self, classes, top_k=(1,)):
        super(BaseEvaluator, self).__init__()
        self.classes = classes
        self.device = torch.device('cpu')

        self.top_k = top_k
        self._init()

    def _init(self):
        self.total_outputs_list = list()
        self.total_targets_list = list()

        self.cate_outputs_dict = dict()
        self.cate_targets_dict = dict()
        for i in range(len(self.classes)):
            key = str(i)
            self.cate_outputs_dict[key] = list()
            self.cate_targets_dict[key] = list()

    def evaluate_train(self, output_dict: dict, targets: torch.Tensor):
        assert isinstance(output_dict, dict) and KEY_OUTPUT in output_dict.keys()

        probs = output_dict[KEY_OUTPUT]
        res, _ = topk_accuracy(probs, targets, top_k=self.top_k)

        acc_dict = dict()
        for i in range(len(self.top_k)):
            acc_dict[f'tok{self.top_k[i]}'] = res[i]
        return acc_dict

    def evaluate_test(self, output_dict: dict, targets: torch.Tensor):
        assert isinstance(output_dict, dict) and KEY_OUTPUT in output_dict.keys()
        probs = output_dict[KEY_OUTPUT]
        outputs = probs.to(device=self.device)
        targets = targets.to(device=self.device)

        self.total_outputs_list.extend(outputs)
        self.total_targets_list.extend(targets)

        for i, target in enumerate(targets):
            key = str(target.item())
            self.cate_outputs_dict[key].append(outputs[i])
            self.cate_targets_dict[key].append(targets[i])

    def get(self):
        assert len(self.total_targets_list) == len(self.total_outputs_list)
        if len(self.total_targets_list) == 0:
            return None, None

        img_length = len(self.total_targets_list)
        result_str = '\ntotal({}) - '.format(img_length)
        topk_list, correct_topk_list = topk_accuracy(torch.stack(self.total_outputs_list),
                                                     torch.stack(self.total_targets_list),
                                                     top_k=self.top_k)
        acc_dict = dict()
        for i in range(len(self.top_k)):
            acc_dict[f"top{self.top_k[i]}"] = topk_list[i]
            result_str += 'top{} acc({}): {:6.3f}%\t'.format(self.top_k[i], correct_topk_list[i], topk_list[i])

        num_len = len(f'({img_length})'.encode('utf-8'))
        max_cls_len = 0
        for cls_name in self.classes:
            if len(cls_name.encode('utf-8')) > max_cls_len:
                max_cls_len = len(cls_name.encode('utf-8'))
        max_cls_len += num_len

        for idx in range(len(self.classes)):
            class_name = self.classes[idx].strip()

            key = str(idx)
            cate_outputs = self.cate_outputs_dict[key]
            result_str += '\n{item1:<4} - {item2:<{length}}\t'.format(item1=idx,
                                                                      item2=f'{class_name}({str(len(cate_outputs))})',
                                                                      length=max_cls_len)
            if len(cate_outputs) == 0:
                for i in range(len(self.top_k)):
                    acc_dict[f"top{self.top_k[i]}"] = 0.
                    result_str += '{item1}: {item2:6.2f}%\t'.format(
                        item1=f'top{self.top_k[i]}',
                        item2=0.)

            else:
                cate_outputs = torch.stack(self.cate_outputs_dict[key])
                cate_targets = torch.stack(self.cate_targets_dict[key])

                topk_list, correct_topk_list = topk_accuracy(cate_outputs, cate_targets, top_k=self.top_k)

                for i in range(len(self.top_k)):
                    acc_dict[f"top{self.top_k[i]}"] = topk_list[i]
                    result_str += '{item1}: {item2:6.2f}%\t'.format(
                        item1=f'top{self.top_k[i]}',
                        item2=topk_list[i]
                    )
        result_str += '\n'

        return result_str, acc_dict

    def clean(self):
        self._init()



# -*- coding: utf-8 -*-

"""
@date: 2020/11/25 下午6:52
@file: dataset.py
@author: zj
@description: 
"""

from yacs.config import CfgNode as CN


def add_config(_C):
    # ---------------------------------------------------------------------------- #
    # DataSet
    # ---------------------------------------------------------------------------- #
    _C.DATASET = CN()
    _C.DATASET.NAME = 'CIFAR100'
    # train data path
    _C.DATASET.TRAIN_ROOT = './data/cifar'
    # test data path
    _C.DATASET.TEST_ROOT = './data/cifar'
    # used for evaluator
    _C.DATASET.TOP_K = (1,)
    # keep rgb data format or not
    _C.DATASET.KEEP_RGB = False



# -*- coding: utf-8 -*-

"""
@date: 2020/11/21 下午2:37
@file: resnet.py
@author: zj
@description: 
"""

from zcls.model import registry
from zcls.model.misc import load_pretrained_weights
from ..base_recognizer import BaseRecognizer

url_map = {
    'resnet18': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/resnet18_imagenet_384d0b7e.pth",
    'resnet34': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/resnet34_imagenet_e4448618.pth",
    'resnet50': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/resnet50_imagenet_025f6510.pth",
    'resnet101': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/resnet101_imagenet_cb164cb4.pth",
    'resnet152': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/resnet152_imagenet_a8be90bf.pth",
    'resnext50_32x4d': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/resnext50_32x4d_imagenet_a21b7284.pth",
    'resnext101_32x8d': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/resnext101_32x8d_imagenet_b4025795.pth",
}


@registry.RECOGNIZER.register('ResNet')
class ResNet(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)

    def _init_weights(self, cfg):
        pretrained_local = cfg.MODEL.RECOGNIZER.PRETRAINED_LOCAL
        pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
        num_classes = cfg.MODEL.HEAD.NUM_CLASSES

        model_name = cfg.MODEL.BACKBONE.ARCH
        assert isinstance(model_name, str)
        if model_name.startswith('repvgg'):
            if len(cfg.MODEL.CONV.ADD_BLOCKS) == 0:
                model_name += '_infer'
            else:
                model_name += '_train'
        load_pretrained_weights(self, model_name,
                                weights_path=None if pretrained_local == '' else pretrained_local,
                                load_fc=pretrained_num_classes == num_classes,
                                verbose=True,
                                url_map=url_map if cfg.MODEL.RECOGNIZER.PRETRAINED_REMOTE else None
                                )



# -*- coding: utf-8 -*-

"""
@date: 2021/6/4 下午5:23
@file: ghostnet.py
@author: zj
@description: 
"""

from zcls.model import registry
from zcls.model.misc import load_pretrained_weights
from ..base_recognizer import BaseRecognizer

url_map = {
    'ghostnet_x1_0': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/ghostnet_x1_0_imagenet_0000e4da.pth",
}


@registry.RECOGNIZER.register('GhostNet')
class GhostNet(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)

    def _init_weights(self, cfg):
        if cfg.MODEL.COMPRESSION.WIDTH_MULTIPLIER == 1.0:
            model_name = 'ghostnet_x1_0'
        else:
            model_name = 'GhostNet'

        pretrained_local = cfg.MODEL.RECOGNIZER.PRETRAINED_LOCAL
        pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
        num_classes = cfg.MODEL.HEAD.NUM_CLASSES

        load_pretrained_weights(self, model_name,
                                weights_path=None if pretrained_local == '' else pretrained_local,
                                load_fc=pretrained_num_classes == num_classes,
                                verbose=True,
                                url_map=url_map if cfg.MODEL.RECOGNIZER.PRETRAINED_REMOTE else None
                                )



# -*- coding: utf-8 -*-

"""
@date: 2020/12/24 下午7:38
@file: shufflenetv1.py
@author: zj
@description: 
"""

from zcls.model import registry
from zcls.model.misc import load_pretrained_weights
from ..base_recognizer import BaseRecognizer

url_map = {
    'mnasnet_b1_0_5': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/mnasnet0_5_imagenet_395ba514.pth",
    'mnasnet_b1_1_0': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/mnasnet1_0_imagenet_51c92e88.pth",
}


@registry.RECOGNIZER.register('MNASNet')
class MNASNet(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)

    def _init_weights(self, cfg):
        pretrained_local = cfg.MODEL.RECOGNIZER.PRETRAINED_LOCAL
        pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
        num_classes = cfg.MODEL.HEAD.NUM_CLASSES

        model_name = cfg.MODEL.BACKBONE.ARCH
        assert isinstance(model_name, str)
        if model_name.startswith('mnasnet_b1'):
            if cfg.MODEL.COMPRESSION.WIDTH_MULTIPLIER == 0.5:
                model_name += '_0_5'
            elif cfg.MODEL.COMPRESSION.WIDTH_MULTIPLIER == 1.0:
                model_name += '_1_0'
            else:
                raise ValueError('No supports values')
        load_pretrained_weights(self, model_name,
                                weights_path=None if pretrained_local == '' else pretrained_local,
                                load_fc=pretrained_num_classes == num_classes,
                                verbose=True,
                                url_map=url_map if cfg.MODEL.RECOGNIZER.PRETRAINED_REMOTE else None
                                )



# -*- coding: utf-8 -*-

"""
@date: 2021/11/7 下午2:22
@file: misc.py
@author: zj
@description: 
"""

import os
import torch

from torch.utils import model_zoo


def load_pretrained_weights(model, model_name, weights_path=None, load_fc=True, verbose=True, url_map=None):
    """Loads pretrained weights from weights path or download using url.

    Args:
        model (Module): The whole model.
        model_name (str): Model name.
        weights_path (None or str):
            str: path to pretrained weights file on the local disk.
            None: use pretrained weights downloaded from the Internet.
        load_fc (bool): Whether to load pretrained weights for fc layer at the end of the model.
        url_map (dict or None): Remote pre-training model path corresponding to each model name
    """
    assert url_map is None or isinstance(url_map, dict), url_map
    assert weights_path is None or os.path.isfile(weights_path), weights_path

    if isinstance(weights_path, str):
        state_dict = torch.load(weights_path)['model']
    elif url_map is not None:
        if model_name not in url_map.keys() or url_map[model_name] == "":
            if verbose:
                print('No pretrained weights for {}'.format(model_name))
            return None
        remote_url = url_map[model_name]
        state_dict = model_zoo.load_url(remote_url)['model']
    else:
        return None

    if load_fc:
        ret = model.load_state_dict(state_dict, strict=False)
        assert not ret.missing_keys, 'Missing keys when loading pretrained weights: {}'.format(ret.missing_keys)
    else:
        missing_keys_list = list()
        if 'head.fc.weight' in state_dict.keys():
            state_dict.pop('head.fc.weight')
            missing_keys_list.append('head.fc.weight')
        if 'head.fc.bias' in state_dict.keys():
            state_dict.pop('head.fc.bias')
            missing_keys_list.append('head.fc.bias')
        if 'head.conv2.weight' in state_dict.keys():
            state_dict.pop('head.conv2.weight')
        if 'head.conv2.bias' in state_dict.keys():
            state_dict.pop('head.conv2.bias')
        ret = model.load_state_dict(state_dict, strict=False)
        assert set(ret.missing_keys) == set(missing_keys_list), \
            'Missing keys when loading pretrained weights: {}'.format(ret.missing_keys)
    # assert not ret.unexpected_keys, 'Missing keys when loading pretrained weights: {}'.format(ret.unexpected_keys)

    if verbose:
        print('Loaded pretrained weights for {}'.format(model_name))



# -*- coding: utf-8 -*-

"""
@date: 2021/4/4 下午2:55
@file: general_dataset.py
@author: zj
@description: 
"""

from torch.utils.data import Dataset
import torchvision.datasets as datasets

from .util import default_converter
from .evaluator.general_evaluator import GeneralEvaluator


class GeneralDataset(Dataset):

    def __init__(self, root, transform=None, target_transform=None, top_k=(1, 5), keep_rgb=False):
        self.data_set = datasets.ImageFolder(root)
        self.classes = self.data_set.classes
        self.root = root
        self.transform = transform
        self.target_transform = target_transform
        self.keep_rgb = keep_rgb
        self._update_evaluator(top_k)

    def __getitem__(self, index: int):
        image, target = self.data_set.__getitem__(index)
        image = default_converter(image, rgb=self.keep_rgb)

        if self.transform is not None:
            image = self.transform(image)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return image, target

    def __len__(self) -> int:
        return self.data_set.__len__()

    def _update_evaluator(self, top_k):
        self.evaluator = GeneralEvaluator(self.classes, top_k=top_k)

    def get_classes(self):
        return self.classes

    def __repr__(self):
        return self.__class__.__name__ + ' (' + self.root + ')'



# -*- coding: utf-8 -*-

"""
@date: 2020/12/24 下午7:38
@file: shufflenetv1.py
@author: zj
@description: 
"""

from zcls.model import registry
from zcls.model.misc import load_pretrained_weights
from ..base_recognizer import BaseRecognizer

"""
G1) Equal channel width minimizes memory access cost (MAC)
G2) Excessive group convolution increases MAC.
G3) Network fragmentation reduces degree of parallelism
G4) Element-wise operations are non-negligible.
"""

url_map = {
    'shufflenet_v2_x0_5': "https://github.com/ZJCV/ZCls/releases/download/v0.14.0/shufflenet_v2_x0_5_imagenet_8fd98bb3.pth",
    'shufflenet_v2_x1_0': "",
    'shufflenet_v2_x1_5': "",
    'shufflenet_v2_x2_0': "",
}


@registry.RECOGNIZER.register('ShuffleNetV2')
class ShuffleNetV2(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)

    def _init_weights(self, cfg):
        pretrained_local = cfg.MODEL.RECOGNIZER.PRETRAINED_LOCAL
        pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
        num_classes = cfg.MODEL.HEAD.NUM_CLASSES
        load_pretrained_weights(self, cfg.MODEL.BACKBONE.ARCH,
                                weights_path=None if pretrained_local == '' else pretrained_local,
                                load_fc=pretrained_num_classes == num_classes,
                                verbose=True,
                                url_map=url_map if cfg.MODEL.RECOGNIZER.PRETRAINED_REMOTE else None
                                )
